/*
  Calisto - An internet Talker
  Copyright (C) 1998-2001 Peter Howkins, Matthew Howkins and Simon Howkins

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,     
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License   
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Email: calisto@calisto.cx

  $Id: plugin2.c,v 1.9 2001/03/08 18:13:43 peter Exp $

 */
static const char rcsid[] = "$Id: plugin2.c,v 1.9 2001/03/08 18:13:43 peter Exp $";

#ifdef USE_PLUGINS
#include <stdio.h>
#include <stdlib.h>

#include "dllist.h"
#include "globals.h"
#include "log.h"
#include "playerdb.h"
#include "plugin.h"
#include "plugin2.h"
#include "pool.h"

#include "./plugins/pstruct.h"

#define CURRENT_PVERSION 1

pool *plugin_pool = 0;
listhead AllPlugins;

void plugin_on_void(unsigned type)
{
  listnode *pNode = AllPlugins.head.next;

  while (LIST_NODE_IS_REAL(pNode)) {
    struct plugin *p = LIST_GET_DATA(pNode, struct plugin *, pluginlink);
    void (*function)(void); 
    if(type == ON_SHUTDOWN) {
      if(p->onshutdown != NULL) {
        function = p->onshutdown;
        function();
      }
    }
    pNode = pNode->next;
  }
}

/* combined on_new, on_login and on_logout */
void plugin_on_char(unsigned type, character *c)
{
  descriptor *d = getdes(c);
  listnode *pNode = AllPlugins.head.next;
  
  while (LIST_NODE_IS_REAL(pNode)) {
    struct plugin *p = LIST_GET_DATA(pNode, struct plugin *, pluginlink);
    void (*function)(character *, void *);

    if(type == ON_NEW) {
      if(p->onnew != NULL) {
        function = p->onnew;
        function(c, ((char *)d) + p->offset);
      }
    } else if(type == ON_LOGIN) {
      if(p->onlogin != NULL) {
        function = p->onlogin;
        function(c, ((char *)d) + p->offset);
      }
    } else if(type == ON_LOGOUT) {
      if(p->onlogout != NULL) {
        function = p->onlogout;
        function(c, ((char *)d) + p->offset);
      }
    }
    pNode = pNode->next;
  }
}

void init_plugins(void)
{
  const char *fname;
  int incoffset = 0; /* the offest to the extra plugin data, incremented */
 
  /* initialise plugins linked list */
  LIST_INIT(AllPlugins);

  log(debug, "Plugin: entered init_plugins()");

  /* Initialise pool-based memory management */
  plugin_pool = pool_create(sizeof(plugin));
  if (!plugin_pool) {
    fprintf(stderr, "Plugin: Not enough memory creating plugin_pool\n");
    log(usage, "Plugin: Not enough memory creating plugin_pool");
    exit(EXIT_FAILURE);
  }

  if(open_plugin_dir() !=0) {
    log(usage, "Plugin: Unable to open Plugin DIR\n");
    return;
  }

  while((fname = next_plugin()) != 0) {
    PLUGIN *cur;

    cur = plugin_open(fname);
    if(cur) {
      struct pheader p;
      unsigned *ver;
      plugin *plugin_new;

      /* check version */
      ver = (unsigned *) plugin_get_symbol(cur, "pversion");
      if(!ver) {
        log(debug, "Plugin '%s' not a Calisto plugin", fname);
        continue;      
      }
      if(*ver != CURRENT_PVERSION) {
        log(debug, "Plugin '%s' wrong version", fname);
        continue;      
      }
 
      plugin_new = pool_malloc(plugin_pool);
      if(!plugin_new) {
        fprintf(stderr, "init_plugins: Out of memory.");
        log(usage, "init_plugins: Out of memory.");
        exit(sizeof(plugin));
      }

      /* Set up the data in the linked list */
      plugin_new->handle = cur;
      p = *(struct pheader *) plugin_get_symbol(cur, "p");

      strcpy(plugin_new->name, p.name);
      strcpy(plugin_new->author, p.author);
      strcpy(plugin_new->helpfile, p.helpfile);
      plugin_new->version = p.version; 

      /* due to oninit needing to be able to setup commands dynamically
         we call the function, then reread the commandlist pointer */
      plugin_new->oninit = p.oninit;
      if(plugin_new->oninit) {
        plugin_new->oninit();
      }
      p = *(struct pheader *) plugin_get_symbol(cur, "p");

      /* back to normal reading of data */
      plugin_new->clist = p.clist;
      if(p.structsize == 0) {
        plugin_new->offset = NO_PLUGIN_EXTRA_DATA;
      } else {
        plugin_new->offset = incoffset + sizeof(descriptor);
      }
      /* setup the functions */
      plugin_new->onfinger = p.onfinger;
      plugin_new->onnew = p.onnew;
      plugin_new->onsave = p.onsave;
      plugin_new->onload = p.onload;
      plugin_new->onlogin = p.onlogin;
      plugin_new->onlogout = p.onlogout;
      plugin_new->onshutdown = p.onshutdown;

      incoffset += p.structsize;
/*      log(debug, "Plugin: loaded plugin '%s'", plugin_new->name); */

      LIST_LINK_NODE_AT_END(AllPlugins, &(plugin_new->pluginlink));
    } else {
      log(debug, "Plugin: Unable to open '%s'", fname);
    }
  }

  close_plugin_dir();

  /* inform main how much memory descriptors will use */
  plugin_extra = incoffset;

  log(debug, "Plugin: leaving init_plugins()");

}

void close_plugins(void)
{
  listnode *pnode = AllPlugins.head.next;

  while(LIST_NODE_IS_REAL(pnode)) {
    plugin *p = LIST_GET_DATA(pnode, plugin *, pluginlink);
  
    plugin_close(p->handle);
    LIST_UNLINK_NODE(&p->pluginlink);
    pool_free(p);  
    pnode = pnode->next;
  }

  if(!LIST_IS_EMPTY(AllPlugins)) 
    log(debug, "close_plugins: error emptying plugins list");


  pool_destroy(plugin_pool);

}

void plugin_reload(character *c)
{
  /* - plan -
     call plugins on_shutdown function
     desruct plugins linked list
     desruct commands linked list
     reboot plugins (inc on_init function)
     reboot commands
     realloc descriptors
     massage AllPlayers list
     call on_login and (on_load or on_new) on all plugins 
  */
  listnode *cNode = AllPlayers.head.next;
  listnode *dNode = AllConns.head.next;

  int dess = 0;  /* number of descriptor/connections */
  int l;
  descriptor *temp, *start;

  send_to_all("[Plugins] Reboot started\n");

  /* call plugins on_shutdown function */
  plugin_on_void(ON_SHUTDOWN);

  /* desruct plugins linked list */
  send_to_char(c, "About to close_plugins()\n");
  close_plugins();  

  /* desruct commands linked list */
  send_to_char(c, "About to free_commands()\n");
  free_commands();

  /* reboot plugins, including called on_init for each plugin */
  send_to_char(c, "About to init_plugins()\n");
  init_plugins();

  /* reboot commands */
  send_to_char(c, "About to init_commands()\n");
  init_commands();

  /* realloc descriptors
      - either 
        - copy descriptors to temporary descriptor array
        - pool_destroy old list
        - pool_create new list
      - or
        - pool_realloc
  */

  /* count number of used descriptors */
  while(LIST_NODE_IS_REAL(dNode)) {
    dess++;
    dNode = dNode->next;
  }

  /* malloc enough space to temporally store the descriptors in */
  start = malloc(sizeof(descriptor) * (dess + 1));
  temp = start;
  if(!temp) {
    /* oh dear, no memory */
    log(usage, "reload plugins failed at memory allocation stage, for temp store");
    fprintf(stderr, "reload plugins failed at memory allocation stage, for temp store\n");
    exit(EXIT_FAILURE);
  }

  /* copy the descriptors to the temporary store, destroy list, and free up
     memory */
  dNode = AllConns.head.next;
  while(LIST_NODE_IS_REAL(dNode)) {
    descriptor *des = LIST_GET_DATA(dNode, descriptor *, descriptorlink); 
    *temp = *des;
    fprintf(stderr, "%d\n", temp->sfd);
    temp++;
    LIST_UNLINK_NODE(dNode);
    pool_free(des);
    dNode = dNode->next;
  }
  temp = NULL;

  /* destroy pool of descriptor list */
  pool_destroy(descriptor_pool);

  /* create new pool of new size */
  descriptor_pool = pool_create(sizeof(descriptor) + plugin_extra);
  if(!descriptor_pool) {
    fprintf(stderr, "Not enough memory to recreate descriptor_pool, in plugin reload\n");
    log(usage, "Not enough memory to recreate descriptor_pool, in plugin reload");
    exit(EXIT_FAILURE);
  }

  /* make new list */
  temp = start;
  for(l=0; l<=dess; l++) {
    descriptor *des = pool_malloc(descriptor_pool);
    if(!des) {
      fprintf(stderr, "pool_malloc faliure in plugin reload\n");
      log(usage, "pool_malloc failure in plugin reload");
      exit(EXIT_FAILURE);
    }
    *des = temp[l];
    LIST_LINK_NODE_AT_END(AllConns, &(des->descriptorlink)); 
  }
  
  /* free up temp store */
  free(start);

  /* paranoia - this should go wrong if we don't do it */
  /* massage AllPlayers list, destruct old list, build from descriptor list */
  cNode = AllPlayers.head.next;
  while(LIST_NODE_IS_REAL(cNode)) {
    LIST_UNLINK_NODE(cNode);
    cNode = cNode->next;
  }  
  dNode = AllConns.head.next;
  while(LIST_NODE_IS_REAL(dNode)) {
    descriptor *des = LIST_GET_DATA(dNode, descriptor *, descriptorlink); 
    if(des->state == STATE_PLAY) {
      LIST_LINK_NODE_AT_END(AllPlayers, &(des->player.characterlink));
    }
    dNode = dNode->next;
  }

  /* call on_login and (on_load or on_new) on all plugins for each player */
  send_to_char(c, "About to init players structs\n");
  cNode = AllPlayers.head.next;
  while(LIST_NODE_IS_REAL(cNode)) {
    character *c = LIST_GET_DATA(cNode, character *, characterlink);
    plugin_on_char(ON_LOGIN, c);
    plugin_on_char(ON_NEW, c);
    load_player(getdes(c), c->name);

    cNode = cNode->next;
  }

  send_to_all("[Plugins] Reboot completed\n");

}

#else /* dont USE_PLUGINS */
static int junk = 0;
#endif /* USE_PLUGINS */

